﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Reflection;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Persistence {
  public sealed class StaticCache {
    private static readonly object locker = new object();

    private readonly Dictionary<Guid, ITransformer> guid2Transformer;
    private readonly Dictionary<ITransformer, Guid> transformer2Guid;
    private readonly Dictionary<Guid, Type> guid2Type;
    private readonly Dictionary<Type, Guid> type2Guid;
    private readonly Dictionary<Type, TypeInfo> typeInfos;
    private readonly Dictionary<Guid, List<MethodInfo>> guid2ConversionMethods;

    internal StaticCache() {
      guid2Transformer = new Dictionary<Guid, ITransformer>();
      transformer2Guid = new Dictionary<ITransformer, Guid>();
      guid2Type = new Dictionary<Guid, Type>();
      type2Guid = new Dictionary<Type, Guid>();
      typeInfos = new Dictionary<Type, TypeInfo>();
      guid2ConversionMethods = new Dictionary<Guid, List<MethodInfo>>();

      foreach (var transformer in ApplicationManager.Manager.GetInstances<ITransformer>())
        RegisterTransformer(transformer);

      RegisterType(new Guid("ECAEA154-6BFF-419F-8BE6-2565E9314825"), typeof(object));
      RegisterType(new Guid("94AD8522-3F55-4580-A6F8-2D2AAEDD4B8C"), typeof(bool));
      RegisterType(new Guid("4A1C0FD5-423D-4F96-AB22-A496578C25AC"), typeof(byte));
      RegisterType(new Guid("C4B00F0B-FED7-439F-B1B2-8A0048B64882"), typeof(sbyte));
      RegisterType(new Guid("9F451811-3DE1-43AB-8B74-D7E03851857B"), typeof(short));
      RegisterType(new Guid("46244D54-0145-49F7-9CF3-9CDB7FB5F240"), typeof(ushort));
      RegisterType(new Guid("1FDDE40C-09E3-491F-8FBB-32BB3C885E9E"), typeof(char));
      RegisterType(new Guid("EE3E0F9C-A5C2-4461-AF36-28BD8F26E6FB"), typeof(int));
      RegisterType(new Guid("69476D18-D285-43E9-BC7C-6CC9E9F2321E"), typeof(uint));
      RegisterType(new Guid("7C7BC5EC-F001-4BA0-9F85-50DCBAA9AE81"), typeof(long));
      RegisterType(new Guid("AC808D5A-63BB-457C-9B92-C9B83DA2B139"), typeof(ulong));
      RegisterType(new Guid("BF22653A-026C-4367-BBBA-2125AECF6C08"), typeof(float));
      RegisterType(new Guid("8B49821A-3ADC-4715-9DB1-08E2F3CFDF15"), typeof(double));
      RegisterType(new Guid("5C6CB86C-81DA-4334-9196-6571C6706240"), typeof(decimal));

      RegisterType(new Guid("7BB386BF-6FD4-443D-A6C1-387096798C67"), typeof(DateTime));
      RegisterType(new Guid("724A2D49-7E7B-455B-BBA9-4214C64E8A21"), typeof(TimeSpan));
      RegisterType(new Guid("ECC12A57-DA8D-43D9-9EC7-FCAC878A4D69"), typeof(Font));
      RegisterType(new Guid("E8348C94-9817-4164-9C98-377689F83F30"), typeof(Bitmap));
      RegisterType(new Guid("05551382-E894-4218-B860-FEE1D92CA07D"), typeof(Nullable<>));
      RegisterType(new Guid("4CC0D44E-65B2-4DF1-A333-30F058CB78CA"), typeof(Tuple<>));
      RegisterType(new Guid("5D451A64-EBD9-411F-A07E-232250B33784"), typeof(Tuple<,>));
      RegisterType(new Guid("EEBD2D26-56CC-45EA-900A-DD29B9F598CE"), typeof(Tuple<,,>));
      RegisterType(new Guid("B8EEEFA7-8576-42DD-937C-AFBB66297B56"), typeof(Tuple<,,,>));
      RegisterType(new Guid("4C3C8E0A-7DCF-4814-B798-AECB86E1E4E3"), typeof(Tuple<,,,,>));
      RegisterType(new Guid("EAA4FF91-B90C-47CB-AC0F-66AB66D41186"), typeof(Tuple<,,,,,>));
      RegisterType(new Guid("EB51B342-826F-4004-B2C6-F2A03A8C35C9"), typeof(Tuple<,,,,,,>));
      RegisterType(new Guid("14138614-45DB-4A0D-9E59-AE9373241D6B"), typeof(Tuple<,,,,,,,>));
      RegisterType(new Guid("4DB2ED2B-D9F2-4695-B555-A2CF42740740"), typeof(Color));
      RegisterType(new Guid("9BC74087-D5C0-4C39-99B4-D7465F478492"), typeof(Point));
      RegisterType(new Guid("E84C326A-7E14-4F28-AEFF-BC16CC671655"), typeof(KeyValuePair<,>));
      RegisterType(new Guid("F0280B55-25E8-4981-B309-D675D081402A"), typeof(string));

      RegisterType(new Guid("9CF55419-439B-4A90-B2ED-8C7F7768EB61"), typeof(Array));
      RegisterType(new Guid("B7313EE2-045F-41C9-AEC9-7B1A8CDBF90F"), typeof(IEnumerable));
      RegisterType(new Guid("D5265599-7380-4A28-AB0D-9F336EF290C5"), typeof(IEnumerable<>));
      RegisterType(new Guid("E451A99B-0CFD-48E6-9BEF-0AE9D1E73628"), typeof(IList));
      RegisterType(new Guid("DB2A785C-5803-42C7-A01F-E1FD845A83A1"), typeof(List<>));
      RegisterType(new Guid("EB98A564-BEDC-458A-9E9B-4BF3CDAFB9BE"), typeof(Stack<>));
      RegisterType(new Guid("D5E3118B-957A-43B5-A740-1BBAA0EAB666"), typeof(Stack));
      RegisterType(new Guid("8CBC5F9F-5671-4805-A690-DD034AEB7932"), typeof(ArrayList));
      RegisterType(new Guid("80CAB5BF-06C2-4DB2-9CF0-EE6E93D40B30"), typeof(Queue));
      RegisterType(new Guid("C528E123-F3DD-4711-83E2-A26E2A9A85E9"), typeof(Queue<>));
      RegisterType(new Guid("7E2F3277-7216-4295-A498-9ACA43527D5B"), typeof(HashSet<>));
      RegisterType(new Guid("E92C35AD-32B1-4F37-B8D2-BE2F5FEB465B"), typeof(Dictionary<,>));
      RegisterType(new Guid("18504ADF-2332-4497-8CA0-B2F00CA2EBA2"), typeof(Type));

      var allTypes = new Dictionary<string, Type>(5000);
      foreach (var asm in AppDomain.CurrentDomain.GetAssemblies()) {
        if (asm.FullName == "mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089") {
          foreach (var t in asm.GetTypes()) {
            if (typeof(IEqualityComparer).IsAssignableFrom(t))
              allTypes.Add(t.FullName, t);

            else if (typeof(IEqualityComparer<>).IsAssignableFrom(t))
              allTypes.Add(t.FullName, t);
          }
        } else {
          foreach (var t in asm.GetTypes()) {
            if (StorableTypeAttribute.IsStorableType(t)) {
              RegisterType(StorableTypeAttribute.GetStorableTypeAttribute(t).Guid, t);
            }
            foreach (var mi in t.GetMethods(BindingFlags.NonPublic | BindingFlags.Static)) {
              if (StorableConversionAttribute.IsStorableConversionMethod(mi)) {
                RegisterStorableConversion(StorableConversionAttribute.GetStorableConversionAttribute(mi).Guid, mi);
              }
            }
          }
        }
      }

      RegisterType(new Guid("67C35B3B-DC23-4929-9BA9-0A0CD5FE17F5"), allTypes["System.StringComparer"]);
      RegisterType(new Guid("C7FDE8F0-EF3F-4A4D-8D87-5559C32A8B90"), allTypes["System.CultureAwareComparer"]);
      RegisterType(new Guid("4AF3A1F6-49CD-4172-8394-FE464889250E"), allTypes["System.CultureAwareRandomizedComparer"]);
      RegisterType(new Guid("6A80741F-066D-460B-B333-4F766A4D4AEF"), allTypes["System.OrdinalComparer"]);
      RegisterType(new Guid("4C7F657C-B69E-4D1A-8812-3254D6219FD2"), allTypes["System.OrdinalRandomizedComparer"]);
      RegisterType(new Guid("0549A3DF-1FD4-487A-B06B-8C572DFFFBEB"), allTypes["System.Security.PermissionTokenKeyComparer"]);
      RegisterType(new Guid("94BB9853-8D25-4DD8-A915-FEF2EFF538BC"), allTypes["System.Resources.FastResourceComparer"]);
      RegisterType(new Guid("65D2DE74-8BDF-4C74-9005-81A2C3991DC5"), allTypes["System.Collections.CompatibleComparer"]);
      RegisterType(new Guid("075827F4-07D9-4E50-9D00-C9FF3E7DCF9A"), allTypes["System.Collections.IEqualityComparer"]);
      RegisterType(new Guid("3DA30D50-2337-4487-AECD-F2DB3ECED834"), allTypes["System.Collections.StructuralEqualityComparer"]);
      RegisterType(new Guid("C8BF5C03-1614-46AC-8730-2F231CAD4232"), allTypes["System.Collections.Generic.EqualityComparer`1"]);
      RegisterType(new Guid("FB247FE0-9203-4196-B4D3-71AB18CD12C5"), allTypes["System.Collections.Generic.GenericEqualityComparer`1"]);
      RegisterType(new Guid("06443E04-FB1D-41E1-BAE2-1AB3728BDCC9"), allTypes["System.Collections.Generic.NullableEqualityComparer`1"]);
      RegisterType(new Guid("D9029E74-C511-4D28-B8C5-D0E299EC45A5"), allTypes["System.Collections.Generic.ObjectEqualityComparer`1"]);
      RegisterType(new Guid("C6D867E2-3CD2-4C59-8BA6-F18F00DDB997"), allTypes["System.Collections.Generic.ByteEqualityComparer"]);
      RegisterType(new Guid("02F8D180-BABE-4073-82E0-7FE77814F53D"), allTypes["System.Collections.Generic.EnumEqualityComparer`1"]);
      RegisterType(new Guid("8A19DA46-FE80-4776-9DB4-B17E6182D104"), allTypes["System.Collections.Generic.SByteEnumEqualityComparer`1"]);
      RegisterType(new Guid("F6F6EFB9-B631-4ACC-9326-F492A6A63011"), allTypes["System.Collections.Generic.ShortEnumEqualityComparer`1"]);
      RegisterType(new Guid("D3D279FE-169E-46B6-93AE-589BB4F24A4D"), allTypes["System.Collections.Generic.LongEnumEqualityComparer`1"]);
      RegisterType(new Guid("BA0AB604-C052-4E08-8F9E-A5D8098F16A6"), allTypes["System.Collections.Generic.RandomizedStringEqualityComparer"]);
      RegisterType(new Guid("00C8C940-63D9-43FF-99BA-9C69301BF043"), allTypes["System.Collections.Generic.RandomizedObjectEqualityComparer"]);
      RegisterType(new Guid("DAF22757-7FCC-49AC-B148-F3DD7E9E0A3B"), allTypes["System.Collections.Generic.IEqualityComparer`1"]);
    }

    public void RegisterTransformer(ITransformer transformer) {
      lock (locker) {
        guid2Transformer.Add(transformer.Guid, transformer);
        transformer2Guid.Add(transformer, transformer.Guid);
      }
    }
    public void RegisterType(Guid guid, Type type) {
      lock (locker) {
        guid2Type.Add(guid, type);
        type2Guid.Add(type, guid);
      }
    }
    public void RegisterStorableConversion(Guid guid, MethodInfo methodInfo) {
      lock (locker) {
        List<MethodInfo> conversionMethods;
        if (!guid2ConversionMethods.TryGetValue(guid, out conversionMethods))
          guid2ConversionMethods.Add(guid, conversionMethods = new List<MethodInfo>());
        conversionMethods.Add(methodInfo);
      }
    }

    // mainly for testing
    public void DeregisterType(Guid guid) {
      lock (locker) {
        type2Guid.Remove(guid2Type[guid]);
        guid2Type.Remove(guid);
      }
    }

    public ITransformer GetTransformer(Guid guid) {
      return guid2Transformer[guid];
    }

    public Guid GetGuid(ITransformer transformer) {
      return transformer2Guid[transformer];
    }

    public Type GetType(Guid guid) {
      return guid2Type[guid];
    }

    public Guid GetGuid(Type type) {
      return type2Guid[type];
    }

    public TypeInfo GetTypeInfo(Type type) {
      lock (locker) {
        TypeInfo typeInfo;
        if (!typeInfos.TryGetValue(type, out typeInfo)) {
          var transformer = guid2Transformer.Values.OrderBy(x => x.Priority).FirstOrDefault(x => x.CanTransformType(type));
          if (transformer == null) throw new PersistenceException("No transformer found for type " + type.AssemblyQualifiedName);
          typeInfo = new TypeInfo(type, transformer);
          typeInfos.Add(type, typeInfo);
        }
        return typeInfo;
      }
    }

    public IEnumerable<MethodInfo> GetStorableConversions(Guid guid) {
      List<MethodInfo> conversionMethods;
      if (guid2ConversionMethods.TryGetValue(guid, out conversionMethods)) return conversionMethods;
      return Enumerable.Empty<MethodInfo>();
    }
  }
}
